@@ -1,21 +1,5 @@
-/*
- * Copyright 2003-2021 The Music Player Daemon Project
- * http://www.musicpd.org
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
- */
+// SPDX-License-Identifier: GPL-2.0-or-later
+// Copyright The Music Player Daemon Project
 
 #include "ProxyDatabasePlugin.hxx"
 #include "db/Interface.hxx"
@@ -32,18 +16,17 @@
 #include "song/UriSongFilter.hxx"
 #include "song/BaseSongFilter.hxx"
 #include "song/TagSongFilter.hxx"
-#include "util/Compiler.h"
 #include "config/Block.hxx"
 #include "tag/Builder.hxx"
 #include "tag/Tag.hxx"
 #include "tag/ParseName.hxx"
-#include "util/ConstBuffer.hxx"
+#include "tag/WithTagBuffer.hxx"
+#include "lib/fmt/RuntimeError.hxx"
 #include "util/RecursiveMap.hxx"
 #include "util/ScopeExit.hxx"
-#include "util/RuntimeError.hxx"
 #include "protocol/Ack.hxx"
-#include "event/SocketMonitor.hxx"
-#include "event/IdleMonitor.hxx"
+#include "event/SocketEvent.hxx"
+#include "event/IdleEvent.hxx"
 #include "Log.hxx"
 
 #include <mpd/client.h>
@@ -66,9 +49,7 @@ public:
 	}
 };
 
-class ProxySong : public LightSong {
-	Tag tag2;
-
+class ProxySong : WithTagBuffer, public LightSong {
 public:
 	explicit ProxySong(const mpd_song *song);
 };
@@ -83,9 +64,15 @@ public:
 	~AllocatedProxySong() {
 		mpd_song_free(song);
 	}
+
+	AllocatedProxySong(const AllocatedProxySong &) = delete;
+	AllocatedProxySong &operator=(const AllocatedProxySong &) = delete;
 };
 
-class ProxyDatabase final : public Database, SocketMonitor, IdleMonitor {
+class ProxyDatabase final : public Database {
+	SocketEvent socket_event;
+	IdleEvent idle_event;
+
 	DatabaseListener &listener;
 
 	const std::string host;
@@ -130,7 +117,7 @@ public:
 		   VisitPlaylist visit_playlist) const override;
 
 	RecursiveMap<std::string> CollectUniqueTags(const DatabaseSelection &selection,
-						    ConstBuffer<TagType> tag_types) const override;
+						    std::span<const TagType> tag_types) const override;
 
 	DatabaseStats GetStats(const DatabaseSelection &selection) const override;
 
@@ -147,11 +134,8 @@ private:
 
 	void Disconnect() noexcept;
 
-	/* virtual methods from SocketMonitor */
-	bool OnSocketReady(unsigned flags) noexcept override;
-
-	/* virtual methods from IdleMonitor */
-	void OnIdle() noexcept override;
+	void OnSocketReady(unsigned flags) noexcept;
+	void OnIdle() noexcept;
 };
 
 static constexpr struct {
@@ -166,9 +150,7 @@ static constexpr struct {
 	{ TAG_NAME, MPD_TAG_NAME },
 	{ TAG_GENRE, MPD_TAG_GENRE },
 	{ TAG_DATE, MPD_TAG_DATE },
-#if LIBMPDCLIENT_CHECK_VERSION(2,12,0)
 	{ TAG_ORIGINAL_DATE, MPD_TAG_ORIGINAL_DATE },
-#endif
 	{ TAG_COMPOSER, MPD_TAG_COMPOSER },
 	{ TAG_PERFORMER, MPD_TAG_PERFORMER },
 	{ TAG_COMMENT, MPD_TAG_COMMENT },
@@ -178,16 +160,32 @@ static constexpr struct {
 	{ TAG_MUSICBRAINZ_ALBUMARTISTID,
 	  MPD_TAG_MUSICBRAINZ_ALBUMARTISTID },
 	{ TAG_MUSICBRAINZ_TRACKID, MPD_TAG_MUSICBRAINZ_TRACKID },
-#if LIBMPDCLIENT_CHECK_VERSION(2,10,0)
 	{ TAG_MUSICBRAINZ_RELEASETRACKID,
 	  MPD_TAG_MUSICBRAINZ_RELEASETRACKID },
-#endif
-#if LIBMPDCLIENT_CHECK_VERSION(2,11,0)
 	{ TAG_ARTIST_SORT, MPD_TAG_ARTIST_SORT },
 	{ TAG_ALBUM_ARTIST_SORT, MPD_TAG_ALBUM_ARTIST_SORT },
-#endif
-#if LIBMPDCLIENT_CHECK_VERSION(2,12,0)
 	{ TAG_ALBUM_SORT, MPD_TAG_ALBUM_SORT },
+#if LIBMPDCLIENT_CHECK_VERSION(2,17,0)
+	{ TAG_WORK, MPD_TAG_WORK },
+	{ TAG_CONDUCTOR, MPD_TAG_CONDUCTOR },
+	{ TAG_LABEL, MPD_TAG_LABEL },
+	{ TAG_GROUPING, MPD_TAG_GROUPING },
+	{ TAG_MUSICBRAINZ_WORKID, MPD_TAG_MUSICBRAINZ_WORKID },
+#endif
+#if LIBMPDCLIENT_CHECK_VERSION(2,20,0)
+	{ TAG_COMPOSERSORT, MPD_TAG_COMPOSER_SORT },
+	{ TAG_ENSEMBLE, MPD_TAG_ENSEMBLE },
+	{ TAG_MOVEMENT, MPD_TAG_MOVEMENT },
+	{ TAG_MOVEMENTNUMBER, MPD_TAG_MOVEMENTNUMBER },
+	{ TAG_LOCATION, MPD_TAG_LOCATION },
+#endif
+#if LIBMPDCLIENT_CHECK_VERSION(2,21,0)
+	{ TAG_MOOD, MPD_TAG_MOOD },
+	{ TAG_MUSICBRAINZ_RELEASEGROUPID,
+	  MPD_TAG_MUSICBRAINZ_RELEASEGROUPID },
+#endif
+#if LIBMPDCLIENT_CHECK_VERSION(2,23,0)
+	{ TAG_SHOWMOVEMENT, MPD_TAG_SHOWMOVEMENT },
 #endif
 	{ TAG_NUM_OF_ITEM_TYPES, MPD_TAG_COUNT }
 };
@@ -207,16 +205,20 @@ Copy(TagBuilder &tag, TagType d_tag,
 }
 
 ProxySong::ProxySong(const mpd_song *song)
-	:LightSong(mpd_song_get_uri(song), tag2)
+	:LightSong(mpd_song_get_uri(song), tag_buffer)
 {
 	const auto _mtime = mpd_song_get_last_modified(song);
 	if (_mtime > 0)
 		mtime = std::chrono::system_clock::from_time_t(_mtime);
 
+#if LIBMPDCLIENT_CHECK_VERSION(2,21,0)
+	if (const auto _added = mpd_song_get_added(song); _added > 0)
+		added = std::chrono::system_clock::from_time_t(_added);
+#endif
+
 	start_time = SongTime::FromS(mpd_song_get_start(song));
 	end_time = SongTime::FromS(mpd_song_get_end(song));
 
-#if LIBMPDCLIENT_CHECK_VERSION(2,15,0)
 	const auto *af = mpd_song_get_audio_format(song);
 	if (af != nullptr) {
 		if (audio_valid_sample_rate(af->sample_rate))
@@ -252,7 +254,6 @@ ProxySong::ProxySong(const mpd_song *song)
 		if (audio_valid_channel_count(af->channels))
 			audio_format.channels = af->channels;
 	}
-#endif
 
 	TagBuilder tag_builder;
 
@@ -263,10 +264,10 @@ ProxySong::ProxySong(const mpd_song *song)
 	for (const auto *i = &tag_table[0]; i->d != TAG_NUM_OF_ITEM_TYPES; ++i)
 		Copy(tag_builder, i->d, song, i->s);
 
-	tag_builder.Commit(tag2);
+	tag_builder.Commit(tag_buffer);
 }
 
-gcc_const
+[[gnu::const]]
 static enum mpd_tag_type
 Convert(TagType tag_type) noexcept
 {
@@ -308,63 +309,16 @@ CheckError(struct mpd_connection *connection)
 		ThrowError(connection);
 }
 
-static bool
-SendConstraints(mpd_connection *connection, const ISongFilter &f)
-{
-	if (auto t = dynamic_cast<const TagSongFilter *>(&f)) {
-		if (t->IsNegated())
-			// TODO implement
-			return true;
-
-		if (t->GetTagType() == TAG_NUM_OF_ITEM_TYPES)
-			return mpd_search_add_any_tag_constraint(connection,
-								 MPD_OPERATOR_DEFAULT,
-								 t->GetValue().c_str());
-
-		const auto tag = Convert(t->GetTagType());
-		if (tag == MPD_TAG_COUNT)
-			return true;
-
-		return mpd_search_add_tag_constraint(connection,
-						     MPD_OPERATOR_DEFAULT,
-						     tag,
-						     t->GetValue().c_str());
-	} else if (auto u = dynamic_cast<const UriSongFilter *>(&f)) {
-		if (u->IsNegated())
-			// TODO implement
-			return true;
-
-		return mpd_search_add_uri_constraint(connection,
-						     MPD_OPERATOR_DEFAULT,
-						     u->GetValue().c_str());
-	} else if (auto b = dynamic_cast<const BaseSongFilter *>(&f)) {
-		return mpd_search_add_base_constraint(connection,
-						      MPD_OPERATOR_DEFAULT,
-						      b->GetValue());
-	} else
-		return true;
-}
-
 static bool
 SendConstraints(mpd_connection *connection, const SongFilter &filter)
 {
-#if LIBMPDCLIENT_CHECK_VERSION(2, 15, 0)
-	if (mpd_connection_cmp_server_version(connection, 0, 21, 0) >= 0)
-		/* with MPD 0.21 (and libmpdclient 2.15), we can pass
-		   arbitrary filters as expression */
-		return mpd_search_add_expression(connection,
-						 filter.ToExpression().c_str());
-#endif
-
-	for (const auto &i : filter.GetItems())
-		if (!SendConstraints(connection, *i))
-			return false;
-
-	return true;
+	return mpd_search_add_expression(connection,
+					 filter.ToExpression().c_str());
 }
 
 static bool
-SendConstraints(mpd_connection *connection, const DatabaseSelection &selection)
+SendConstraints(mpd_connection *connection, const DatabaseSelection &selection,
+		const RangeArg &window)
 {
 	if (!selection.uri.empty() &&
 	    !mpd_search_add_base_constraint(connection,
@@ -376,16 +330,18 @@ SendConstraints(mpd_connection *connection, const DatabaseSelection &selection)
 	    !SendConstraints(connection, *selection.filter))
 		return false;
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 11, 0)
-	if (selection.sort != TAG_NUM_OF_ITEM_TYPES &&
-	    mpd_connection_cmp_server_version(connection, 0, 21, 0) >= 0) {
-#if LIBMPDCLIENT_CHECK_VERSION(2, 15, 0)
+	if (selection.sort != TAG_NUM_OF_ITEM_TYPES) {
 		if (selection.sort == SORT_TAG_LAST_MODIFIED) {
 			if (!mpd_search_add_sort_name(connection, "Last-Modified",
 						      selection.descending))
 				return false;
-		} else {
+#if LIBMPDCLIENT_CHECK_VERSION(2,21,0)
+		} else if (selection.sort == SORT_TAG_ADDED) {
+			if (!mpd_search_add_sort_name(connection, "Added",
+						      selection.descending))
+				return false;
 #endif
+		} else {
 			const auto sort = Convert(selection.sort);
 			/* if this is an unsupported tag, the sort
 			   will be done later by class
@@ -394,19 +350,12 @@ SendConstraints(mpd_connection *connection, const DatabaseSelection &selection)
 			    !mpd_search_add_sort_tag(connection, sort,
 						     selection.descending))
 				return false;
-#if LIBMPDCLIENT_CHECK_VERSION(2, 15, 0)
 		}
-#endif
 	}
-#endif
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 10, 0)
-	if (selection.window != RangeArg::All() &&
-	    mpd_connection_cmp_server_version(connection, 0, 20, 0) >= 0 &&
-	    !mpd_search_add_window(connection, selection.window.start,
-				   selection.window.end))
+	if (window != RangeArg::All() &&
+	    !mpd_search_add_window(connection, window.start, window.end))
 		return false;
-#endif
 
 	return true;
 }
@@ -416,28 +365,21 @@ SendGroup(mpd_connection *connection, TagType group)
 {
 	assert(group != TAG_NUM_OF_ITEM_TYPES);
 
-#if LIBMPDCLIENT_CHECK_VERSION(2,12,0)
 	const auto tag = Convert(group);
 	if (tag == MPD_TAG_COUNT)
 		throw std::runtime_error("Unsupported tag");
 
 	return mpd_search_add_group_tag(connection, tag);
-#else
-	(void)connection;
-	(void)group;
-
-	throw std::runtime_error("Grouping requires libmpdclient 2.12");
-#endif
 }
 
 static bool
-SendGroup(mpd_connection *connection, ConstBuffer<TagType> group)
+SendGroup(mpd_connection *connection, std::span<const TagType> group)
 {
 	while (!group.empty()) {
 		if (!SendGroup(connection, group.back()))
 		    return false;
 
-		group.pop_back();
+		group = group.first(group.size() - 1);
 	}
 
 	return true;
@@ -446,7 +388,8 @@ SendGroup(mpd_connection *connection, ConstBuffer<TagType> group)
 ProxyDatabase::ProxyDatabase(EventLoop &_loop, DatabaseListener &_listener,
 			     const ConfigBlock &block)
 	:Database(proxy_db_plugin),
-	 SocketMonitor(_loop), IdleMonitor(_loop),
+	 socket_event(_loop, BIND_THIS_METHOD(OnSocketReady)),
+	 idle_event(_loop, BIND_THIS_METHOD(OnIdle)),
 	 listener(_listener),
 	 host(block.GetBlockValue("host", "")),
 	 password(block.GetBlockValue("password", "")),
@@ -495,12 +438,12 @@ ProxyDatabase::Connect()
 	try {
 		CheckError(connection);
 
-		if (mpd_connection_cmp_server_version(connection, 0, 19, 0) < 0) {
+		if (mpd_connection_cmp_server_version(connection, 0, 21, 0) < 0) {
 			const unsigned *version =
 				mpd_connection_get_server_version(connection);
-			throw FormatRuntimeError("Connect to MPD %u.%u.%u, but this "
-						 "plugin requires at least version 0.19",
-						 version[0], version[1], version[2]);
+			throw FmtRuntimeError("Connect to MPD {}.{}.{}, but this "
+					      "plugin requires at least version 0.21",
+					      version[0], version[1], version[2]);
 		}
 
 		if (!password.empty() &&
@@ -512,22 +455,17 @@ ProxyDatabase::Connect()
 
 		std::throw_with_nested(host.empty()
 				       ? std::runtime_error("Failed to connect to remote MPD")
-				       : FormatRuntimeError("Failed to connect to remote MPD '%s'",
-							    host.c_str()));
+				       : FmtRuntimeError("Failed to connect to remote MPD {:?}",
+							 host));
 	}
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 10, 0)
 	mpd_connection_set_keepalive(connection, keepalive);
-#else
-	// suppress -Wunused-private-field
-	(void)keepalive;
-#endif
 
 	idle_received = ~0U;
 	is_idle = false;
 
-	SocketMonitor::Open(SocketDescriptor(mpd_async_get_fd(mpd_connection_get_async(connection))));
-	IdleMonitor::Schedule();
+	socket_event.Open(SocketDescriptor(mpd_async_get_fd(mpd_connection_get_async(connection))));
+	idle_event.Schedule();
 }
 
 void
@@ -554,7 +492,7 @@ ProxyDatabase::CheckConnection()
 
 		idle_received |= idle;
 		is_idle = false;
-		IdleMonitor::Schedule();
+		idle_event.Schedule();
 	}
 }
 
@@ -572,23 +510,23 @@ ProxyDatabase::Disconnect() noexcept
 {
 	assert(connection != nullptr);
 
-	IdleMonitor::Cancel();
-	SocketMonitor::Steal();
+	idle_event.Cancel();
+	socket_event.ReleaseSocket();
 
 	mpd_connection_free(connection);
 	connection = nullptr;
 }
 
-bool
+void
 ProxyDatabase::OnSocketReady([[maybe_unused]] unsigned flags) noexcept
 {
 	assert(connection != nullptr);
 
 	if (!is_idle) {
 		// TODO: can this happen?
-		IdleMonitor::Schedule();
-		SocketMonitor::Cancel();
-		return true;
+		idle_event.Schedule();
+		socket_event.Cancel();
+		return;
 	}
 
 	auto idle = (unsigned)mpd_recv_idle(connection, false);
@@ -598,16 +536,15 @@ ProxyDatabase::OnSocketReady([[maybe_unused]] unsigned flags) noexcept
 		} catch (...) {
 			LogError(std::current_exception());
 			Disconnect();
-			return false;
+			return;
 		}
 	}
 
 	/* let OnIdle() handle this */
 	idle_received |= idle;
 	is_idle = false;
-	IdleMonitor::Schedule();
-	SocketMonitor::Cancel();
-	return true;
+	idle_event.Schedule();
+	socket_event.Cancel();
 }
 
 void
@@ -635,14 +572,14 @@ ProxyDatabase::OnIdle() noexcept
 			LogError(std::current_exception());
 		}
 
-		SocketMonitor::Steal();
+		socket_event.ReleaseSocket();
 		mpd_connection_free(connection);
 		connection = nullptr;
 		return;
 	}
 
 	is_idle = true;
-	SocketMonitor::ScheduleRead();
+	socket_event.ScheduleRead();
 }
 
 const LightSong *
@@ -707,7 +644,7 @@ Visit(struct mpd_connection *connection,
 		      visit_directory, visit_song, visit_playlist);
 }
 
-gcc_pure
+[[gnu::pure]]
 static bool
 Match(const SongFilter *filter, const LightSong &song) noexcept
 {
@@ -828,26 +765,48 @@ try {
 	const bool exact = selection.filter == nullptr ||
 		!selection.filter->HasFoldCase();
 
-	if (!mpd_search_db_songs(connection, exact) ||
-	    !SendConstraints(connection, selection) ||
-	    !mpd_search_commit(connection))
-		ThrowError(connection);
+	/* request only this number of songs at a time to avoid
+	   blowing the server's max_output_buffer_size limit */
+	constexpr unsigned LIMIT = 4096;
 
-	while (auto *song = mpd_recv_song(connection)) {
-		AllocatedProxySong song2(song);
+	auto remaining_window = selection.window;
 
-		if (Match(selection.filter, song2)) {
-			try {
-				visit_song(song2);
-			} catch (...) {
-				mpd_response_finish(connection);
-				throw;
+	while (remaining_window.start < remaining_window.end) {
+		auto window = remaining_window;
+		if (window.end - window.start > LIMIT)
+			window.end = window.start + LIMIT;
+
+		if (!mpd_search_db_songs(connection, exact) ||
+		    !SendConstraints(connection, selection, window) ||
+		    !mpd_search_commit(connection))
+			ThrowError(connection);
+
+		while (auto *song = mpd_recv_song(connection)) {
+			++window.start;
+
+			AllocatedProxySong song2(song);
+
+			if (Match(selection.filter, song2)) {
+				try {
+					visit_song(song2);
+				} catch (...) {
+					mpd_response_finish(connection);
+					throw;
+				}
 			}
 		}
-	}
 
-	if (!mpd_response_finish(connection))
-		ThrowError(connection);
+		if (!mpd_response_finish(connection))
+			ThrowError(connection);
+
+		if (window.start != window.end)
+			/* the other MPD has given us less than we
+			   requested - this means there's no more
+			   data */
+			break;
+
+		remaining_window.start = window.end;
+	}
 } catch (...) {
 	if (connection != nullptr)
 		mpd_search_cancel(connection);
@@ -855,117 +814,40 @@ try {
 	throw;
 }
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 10, 0)
-
-gcc_pure
+[[gnu::pure]]
 static bool
-IsFilterSupported(const ISongFilter &f) noexcept
+IsSortSupported(TagType tag_type) noexcept
 {
-	if (auto t = dynamic_cast<const TagSongFilter *>(&f)) {
-		if (t->IsNegated())
-			// TODO implement
-			return false;
-
-		if (t->GetTagType() == TAG_NUM_OF_ITEM_TYPES)
-			return true;
-
-		const auto tag = Convert(t->GetTagType());
-		return tag != MPD_TAG_COUNT;
-	} else if (auto u = dynamic_cast<const UriSongFilter *>(&f)) {
-		if (u->IsNegated())
-			// TODO implement
-			return false;
-
-		return false;
-	} else if (dynamic_cast<const BaseSongFilter *>(&f)) {
-		return true;
-	} else
-		return false;
-}
-
-gcc_pure
-static bool
-IsFilterFullySupported(const SongFilter &filter,
-		       const struct mpd_connection *connection) noexcept
-{
-#if LIBMPDCLIENT_CHECK_VERSION(2, 15, 0)
-	if (mpd_connection_cmp_server_version(connection, 0, 21, 0) >= 0)
-		/* with MPD 0.21 (and libmpdclient 2.15), we can pass
-		   arbitrary filters as expression */
-		return true;
-#else
-	(void)connection;
-#endif
-
-	for (const auto &i : filter.GetItems())
-		if (!IsFilterSupported(*i))
-			return false;
-
-	return true;
-}
-
-gcc_pure
-static bool
-IsFilterFullySupported(const SongFilter *filter,
-		       const struct mpd_connection *connection) noexcept
-{
-	return filter == nullptr ||
-		IsFilterFullySupported(*filter, connection);
-}
-
-#endif
-
-#if LIBMPDCLIENT_CHECK_VERSION(2, 11, 0)
-
-gcc_pure
-static bool
-IsSortSupported(TagType tag_type,
-		const struct mpd_connection *connection) noexcept
-{
-	if (mpd_connection_cmp_server_version(connection, 0, 21, 0) < 0)
-		/* sorting requires MPD 0.21 */
-		return false;
-
 	if (tag_type == TagType(SORT_TAG_LAST_MODIFIED)) {
-		/* sort "Last-Modified" requires libmpdclient 2.15 for
-		   mpd_search_add_sort_name() */
-#if LIBMPDCLIENT_CHECK_VERSION(2, 15, 0)
 		return true;
-#else
-		return false;
-#endif
 	}
 
+#if LIBMPDCLIENT_CHECK_VERSION(2,21,0)
+	if (tag_type == TagType(SORT_TAG_ADDED)) {
+		return true;
+	}
+#endif
+
 	return Convert(tag_type) != MPD_TAG_COUNT;
 }
 
-#endif
-
-gcc_pure
+[[gnu::pure]]
 static DatabaseSelection
-CheckSelection(DatabaseSelection selection,
-	       struct mpd_connection *connection) noexcept
+CheckSelection(DatabaseSelection selection) noexcept
 {
 	selection.uri.clear();
 	selection.filter = nullptr;
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 11, 0)
 	if (selection.sort != TAG_NUM_OF_ITEM_TYPES &&
-	    IsSortSupported(selection.sort, connection))
+	    IsSortSupported(selection.sort))
 		/* we can forward the "sort" parameter to the other
 		   MPD */
 		selection.sort = TAG_NUM_OF_ITEM_TYPES;
-#endif
 
-#if LIBMPDCLIENT_CHECK_VERSION(2, 10, 0)
-	if (selection.window != RangeArg::All() &&
-	    IsFilterFullySupported(selection.filter, connection))
+	if (selection.window != RangeArg::All())
 		/* we can forward the "window" parameter to the other
 		   MPD */
 		selection.window = RangeArg::All();
-#else
-	(void)connection;
-#endif
 
 	return selection;
 }
@@ -979,11 +861,11 @@ ProxyDatabase::Visit(const DatabaseSelection &selection,
 	// TODO: eliminate the const_cast
 	const_cast<ProxyDatabase *>(this)->EnsureConnected();
 
-	DatabaseVisitorHelper helper(CheckSelection(selection, connection),
+	DatabaseVisitorHelper helper(CheckSelection(selection),
 				     visit_song);
 
 	if (!visit_directory && !visit_playlist && selection.recursive &&
-	    !selection.IsEmpty()) {
+	    selection.IsFiltered()) {
 		/* this optimized code path can only be used under
 		   certain conditions */
 		::SearchSongs(connection, selection, visit_song);
@@ -1001,7 +883,7 @@ ProxyDatabase::Visit(const DatabaseSelection &selection,
 
 RecursiveMap<std::string>
 ProxyDatabase::CollectUniqueTags(const DatabaseSelection &selection,
-				 ConstBuffer<TagType> tag_types) const
+				 std::span<const TagType> tag_types) const
 try {
 	// TODO: eliminate the const_cast
 	const_cast<ProxyDatabase *>(this)->EnsureConnected();
@@ -1010,11 +892,10 @@ try {
 	if (tag_type2 == MPD_TAG_COUNT)
 		throw std::runtime_error("Unsupported tag");
 
-	auto group = tag_types;
-	group.pop_back();
+	const auto group = tag_types.first(tag_types.size() - 1);
 
 	if (!mpd_search_db_tags(connection, tag_type2) ||
-	    !SendConstraints(connection, selection) ||
+	    !SendConstraints(connection, selection, selection.window) ||
 	    !SendGroup(connection, group))
 		ThrowError(connection);
 
